"""
Read data sent by Micromed through TCP.
"""
import logging

import click
from pathlib import Path
from micromed_io.trc import MicromedTRC


@click.command(context_settings=dict(max_content_width=120))
@click.option(
    "--dirpath",
    default="./",
    type=str,
    required=False,
    help="the relative or absolute path containing the TRC files to rename",
    show_default=True,
)
@click.option(
    "--format",
    default="%Y%m%d-%H%M%S",
    type=str,
    required=False,
    help="the datetime format to use - must be compliant with python strftime",
    show_default=True,
)
def run(dirpath: str = "./", format: str = "%Y%m%d-%H%M%S") -> None:
    """Rename the standard TRC files to include the recording date in the filename"""
    logging.basicConfig(
        level=0,
        format=(
            "[%(asctime)s - %(filename)s:%(lineno)d]\t\t%(levelname)s\t\t%(message)s"
        ),
    )
    try:
        for file in Path(dirpath).glob("*.TRC"):
            mmtrc = MicromedTRC(file)
            logging.info(
                f"Renaming to: {file.name[:-4]}__{mmtrc.get_header().recording_date.strftime(format)}"
            )
            file.rename(
                file.parent
                / f"{file.name[:-4]}__{mmtrc.get_header().recording_date.strftime(format)}.TRC"
            )

    except Exception as e:
        logging.error(e)


if __name__ == "__main__":
    run()



"""Micromed TCP module."""

import sys
from enum import IntEnum
import struct
import numpy as np

import logging

from micromed_io.in_out import MicromedIO


class MicromedPacketType(IntEnum):
    """Micromed packet type."""

    HEADER = 0
    EEG_DATA = 1
    NOTE = 2
    MARKER = 3


def get_tcp_header(packet_type: MicromedPacketType, size: int) -> bytearray:
    """Build the Micromed TCP header

    Parameters
    ----------
    packet_type : MicromedPacketType
        The packet type to be sent
    size : int
        The size of the packet to be sent

    Returns
    -------
    bytearray
        The TCP header sent right before any packet
    """
    tcp_header = bytearray(b"MICM")
    tcp_header.extend((packet_type).to_bytes(2, byteorder="little"))
    tcp_header.extend(size.to_bytes(4, byteorder="little"))
    return tcp_header


def decode_tcp_header_packet(packet: bytearray):
    """Decode the Micromed tcp header packet.

    Parameters
    ----------
    packet : bytearray
        The tcp packet to decode.

    Returns
    -------
    bool
        True if decoding was successful. Else False.
    """
    packet_type = None
    next_packet_size = None
    if len(packet) > 0:
        # Check that the packet is not corrupted
        if packet[:4].decode() == "MICM":
            packet_type = int.from_bytes(packet[4:6], "little")
            next_packet_size = int.from_bytes(packet[6:10], "little")
        else:
            logging.warning(f"Wrong header packet: {packet.decode()}")
    else:
        logging.warning(f"header empty packet: [{packet.decode()}]")
    return packet_type, next_packet_size


def decode_tcp_marker_packet(packet: bytearray):
    """Decode the Micromed tcp markers packet.

    Parameters
    ----------
    packet : bytearray
        The tcp packet to decode.

    Returns
    -------
    tuple
        trigger_sample: the sample when the marker is received
        trigger_value: trigger value
    """

    trigger_sample = int.from_bytes(packet[:4], byteorder="little")
    trigger_value = int.from_bytes(packet[4:6], byteorder="little")
    return trigger_sample, trigger_value


def decode_tcp_note_packet(packet: bytearray):
    """Decode the Micromed tcp markers packet.

    Parameters
    ----------
    packet : bytearray
        The tcp packet to decode.

    Returns
    -------
    tuple
        trigger_sample: the sample when the marker is received
        trigger_value: trigger value
    """

    note_sample = struct.unpack("<I", packet[:4])
    note_value = packet[4:].decode(encoding="utf-8").rstrip("\x00")
    return note_sample, note_value


def encode_note_packet(sample: np.uint32, note: str) -> bytearray:
    """Encode a note packet with sample and text

    Parameters
    ----------
    sample : np.uint32
        The sample when note occured
    note : str
        The text

    Returns
    -------
    bytearray
        The encoded packet
    """
    packet = struct.pack("<I", sample)
    packet += struct.pack("<40s", (note).encode(encoding="utf-8"))
    return packet


def encode_marker_packet(sample: np.uint32, marker: np.uint16) -> bytearray:
    """Encode a marker packet with sample and value

    Parameters
    ----------
    sample : np.uint32
        The sample when marker occured
    marker : np.uint16
        The marker value

    Returns
    -------
    bytearray
        The encoded packet
    """
    packet = struct.pack("<I", sample)
    packet += struct.pack("<H", marker)
    return packet



"""
Read data sent by Micromed through TCP.
"""
import logging
import socket
from datetime import datetime

import click
import numpy as np
import pylsl

from micromed_io.in_out import MicromedHeader, MicromedIO
import micromed_io.tcp as mmio_tcp


def recvall(sock, n):
    # Helper function to recv n bytes or return None if EOF is hit
    data = bytearray()
    while len(data) < n:
        packet = sock.recv(n - len(data))
        if not packet:
            return None
        data.extend(packet)
    return data


def init_lsl(
    header: MicromedHeader,
    lsl_eeg_name: str = "Micromed",
    lsl_eeg_type: str = "EEG",
    lsl_eeg_source_id: str = "micromed_eeg",
    lsl_markers_name: str = "Micromed_Markers",
    lsl_markers_type: str = "Markers",
    lsl_markers_source_id: str = "micromed_markers",
    lsl_notes_name: str = "Micromed_Notes",
    lsl_notes_type: str = "Markers",
    lsl_notes_source_id: str = "micromed_notes",
) -> (pylsl.StreamOutlet, pylsl.StreamOutlet):
    """Initializes the Labstreaming Layer outlet with the Micromed Header information"""
    # data outlet
    info = pylsl.StreamInfo(
        name=lsl_eeg_name,
        type=lsl_eeg_type,
        channel_count=int(header.nb_of_channels),
        nominal_srate=int(header.min_sampling_rate),
        channel_format="float32",
        source_id=lsl_eeg_source_id,
    )

    # Create EEG data outlet
    info.desc().append_child_value("manufacturer", "Micromed")
    channels = info.desc().append_child("channels")
    for _, ch_name in enumerate(header.ch_names):
        channels.append_child("channel").append_child_value(
            "label", ch_name
        ).append_child_value("unit", "microvolts").append_child_value("type", "EEG")
    eeg_outlet = pylsl.StreamOutlet(info)

    # Create MARKERS outlet
    info_markers = pylsl.StreamInfo(
        name=lsl_markers_name,
        type=lsl_markers_type,
        channel_count=2,  # TODO: proper doc - [sample, value]
        nominal_srate=0,
        channel_format="int32",
        source_id=lsl_markers_source_id,
    )
    info_markers.desc().append_child_value("manufacturer", "Micromed_markers")
    channels = info_markers.desc().append_child("channels")
    for label in ["marker_sample", "marker_value"]:
        ch = channels.append_child("channel")
        ch.append_child_value("label", label)
        ch.append_child_value("type", "Marker")
    markers_outlet = pylsl.StreamOutlet(info_markers)

    # Create NOTES outlet
    info_notes = pylsl.StreamInfo(
        name=lsl_notes_name,
        type=lsl_notes_type,
        channel_count=2,  # TODO: proper doc - [sample, value]
        nominal_srate=0,
        channel_format="string",
        source_id=lsl_notes_source_id,
    )
    info_notes.desc().append_child_value("manufacturer", "Micromed_notes")
    channels = info_notes.desc().append_child("channels")
    for label in ["note_sample", "note_value"]:
        ch = channels.append_child("channel")
        ch.append_child_value("label", label)
        ch.append_child_value("type", "Marker")
    notes_outlet = pylsl.StreamOutlet(info_notes)

    # ready to parse
    logging.info("LSL server ready.")

    return eeg_outlet, markers_outlet, notes_outlet


@click.command(context_settings=dict(max_content_width=120))
@click.option(
    "--address",
    "-a",
    default="localhost",
    type=str,
    required=False,
    help="the TCP address to use for the server (your IP)",
    show_default=True,
)
@click.option(
    "--port",
    "-p",
    default=5123,
    type=int,
    required=False,
    help="The TCP port number to use",
    show_default=True,
)
@click.option(
    "--verbosity",
    "-v",
    default="1",
    type=click.Choice(["0", "1", "2"]),
    required=False,
    help="Increase output verbosity",
    show_default=True,
)
@click.option(
    "--lsl-eeg-name",
    default="Micromed",
    type=str,
    required=False,
    help="the LSL stream name for the eeg",
    show_default=True,
)
@click.option(
    "--lsl-eeg-type",
    default="EEG",
    type=str,
    required=False,
    help="the LSL stream type for the eeg",
    show_default=True,
)
@click.option(
    "--lsl-eeg-source-id",
    default="micromed_eeg",
    type=str,
    required=False,
    help="the LSL stream source id for the eeg",
    show_default=True,
)
@click.option(
    "--lsl-markers-name",
    default="Micromed_Markers",
    type=str,
    required=False,
    help="the LSL stream name for the markers",
    show_default=True,
)
@click.option(
    "--lsl-markers-type",
    default="Markers",
    type=str,
    required=False,
    help="the LSL stream type for the markers",
    show_default=True,
)
@click.option(
    "--lsl-markers-source-id",
    default="micromed_marker",
    type=str,
    required=False,
    help="the LSL stream source id for the markers",
    show_default=True,
)
@click.option(
    "--lsl-notes-name",
    default="Micromed_Notes",
    type=str,
    required=False,
    help="the LSL stream name for the notes",
    show_default=True,
)
@click.option(
    "--lsl-notes-type",
    default="Markers",
    type=str,
    required=False,
    help="the LSL stream type for the notes",
    show_default=True,
)
@click.option(
    "--lsl-notes-source-id",
    default="micromed_notes",
    type=str,
    required=False,
    help="the LSL stream source id for the notes",
    show_default=True,
)
def run(
    address: str = "localhost",
    port: int = 5123,
    lsl_eeg_name: str = "Micromed",
    lsl_eeg_type: str = "EEG",
    lsl_eeg_source_id: str = "micromed_eeg",
    lsl_markers_name: str = "Micromed_Markers",
    lsl_markers_type: str = "Markers",
    lsl_markers_source_id: str = "micromed_markers",
    lsl_notes_name: str = "Micromed_Notes",
    lsl_notes_type: str = "Markers",
    lsl_notes_source_id: str = "micromed_notes",
    verbosity: int = 1,
) -> None:
    logging.basicConfig(
        level=0,
        format=(
            "[%(asctime)s - %(filename)s:%(lineno)d]\t\t%(levelname)s\t\t%(message)s"
        ),
    )

    verbosity = int(verbosity)  # because of click choice...
    previous_eeg_packet_time = datetime.now()

    while True:
        # Create a TCP/IP socket
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        # Bind the socket to the port
        server_address = (address, port)
        if verbosity >= 1:
            logging.info("starting up on %s port %s" % server_address)
        sock.bind(server_address)
        sock.listen(1)

        sock.settimeout(5)
        connection = None

        logging.info("Waiting for a connection...")
        while connection is None:
            # Wait for a connection
            try:
                connection, _ = sock.accept()
                if verbosity >= 1:
                    logging.info("Connection from %s port %s" % server_address)
            except Exception as e:
                logging.warning(e)

        sock.settimeout(None)

        lsl_eeg_outlet = None
        lsl_markers_outlet = None
        # create micromed
        micromed_io = MicromedIO()
        try:
            while True:
                header = connection.recv(10)  # 10 is enoughbut more is fine too
                b_header = bytearray(header)
                packet_type, next_packet_size = mmio_tcp.decode_tcp_header_packet(
                    b_header
                )

                if packet_type is not None:
                    data = recvall(connection, next_packet_size)
                    b_data = bytearray(data)

                    if packet_type == mmio_tcp.MicromedPacketType.HEADER:
                        micromed_io.decode_data_header_packet(b_data)

                        logging.info("Got Micromed header. Init LSL stream...")
                        if verbosity >= 1:
                            logging.debug(
                                f"n_channels={micromed_io.micromed_header.nb_of_channels}, "
                                + f"sfreq={micromed_io.sfreq}, "
                                + f"first 10 ch_names: {micromed_io.micromed_header.ch_names[:10]}"
                            )
                        lsl_eeg_outlet, lsl_markers_outlet, lsl_notes_outlet = init_lsl(
                            micromed_io.micromed_header,
                            lsl_eeg_name=lsl_eeg_name,
                            lsl_eeg_type=lsl_eeg_type,
                            lsl_eeg_source_id=lsl_eeg_source_id,
                            lsl_markers_name=lsl_markers_name,
                            lsl_markers_type=lsl_markers_type,
                            lsl_markers_source_id=lsl_markers_source_id,
                            lsl_notes_name=lsl_notes_name,
                            lsl_notes_type=lsl_notes_type,
                            lsl_notes_source_id=lsl_notes_source_id,
                        )

                    elif packet_type == mmio_tcp.MicromedPacketType.EEG_DATA:
                        if not micromed_io.decode_data_eeg_packet(b_data):
                            logging.error("Error in EEG data packet")
                        # forward to lsl
                        if lsl_eeg_outlet is not None:
                            # send to lsl
                            lsl_eeg_outlet.push_chunk(
                                np.ascontiguousarray(
                                    micromed_io.current_data_eeg.T.astype("float32")
                                )
                            )

                            # log only
                            if (
                                verbosity >= 1
                                and (
                                    datetime.now() - previous_eeg_packet_time
                                ).total_seconds()
                                > 1
                            ):  # don't bother user too much - wait 2 sec between 2 logs
                                previous_eeg_packet_time = datetime.now()
                                logging.debug(
                                    f"Receiving TCP packet - LSL Sending chunk of size {micromed_io.current_data_eeg.T.shape}"
                                )
                        # log only
                        if verbosity >= 2:
                            logging.info(
                                f"Received EEG data: {micromed_io.current_data_eeg}"
                            )

                    elif packet_type == mmio_tcp.MicromedPacketType.NOTE:
                        note_sample, note_value = mmio_tcp.decode_tcp_note_packet(
                            b_data
                        )
                        lsl_notes_outlet.push_sample([str(note_sample), note_value])
                        if verbosity >= 1:
                            logging.info(
                                f"Received note: sample={note_sample} ,value={note_value}"
                            )

                    elif packet_type == mmio_tcp.MicromedPacketType.MARKER:
                        marker_sample, marker_value = mmio_tcp.decode_tcp_marker_packet(
                            b_data
                        )
                        lsl_markers_outlet.push_sample([marker_sample, marker_value])
                        if verbosity >= 1:
                            logging.info(
                                f"Received marker: sample={marker_sample} ,value={marker_value}"
                            )
                    else:
                        raise ValueError(
                            f"ERROR in packet ! Unknown tcp_data_type: {packet_type}"
                        )
                else:
                    raise ValueError("ERROR: Wrong header. Skipping data")

        except Exception as e:
            logging.error(e)

        finally:
            # Clean up the connection
            if connection is not None:
                if verbosity >= 1:
                    logging.info("Closing the connection")
                connection.close()
            sock.close()


if __name__ == "__main__":
    run()



"""
Emulate from a TRC file Micromed sending data through TCP.

Note: The server must be launched before.

HOW TO USE:
> python emulate_trc_tcpip.py --help
> python emulate_trc_tcpip.py --file=../data/sample.TRC
"""
import logging
import socket
import numpy as np
import time
from datetime import datetime
from pathlib import Path
import click

from micromed_io.in_out import MicromedIO
import micromed_io.tcp as mmio_tcp


@click.command(context_settings=dict(max_content_width=120))
@click.option("--file", "-f", type=str, required=True, help="the TRC file to emulate")
@click.option(
    "--address",
    "-a",
    default="localhost",
    type=str,
    required=False,
    help="the TCP address to connect to",
    show_default=True,
)
@click.option(
    "--port",
    "-p",
    default=5123,
    type=int,
    required=False,
    help="The TCP port number to connect to",
    show_default=True,
)
@click.option(
    "--packet-time",
    "-pt",
    default=256,
    type=int,
    required=False,
    help="The time (in ms) of data to send",
    show_default=True,
)
@click.option(
    "--verbosity",
    "-v",
    default="1",
    type=click.Choice(["0", "1", "2"]),
    required=False,
    help="Increase output verbosity",
    show_default=True,
)
def run(
    file: str,
    address: str = "localhost",
    port: int = 5123,
    packet_time: int = 256,
    verbosity: int = 1,
) -> None:
    """Emulate a Micromed TCP client based on a TRC file"""
    logging.basicConfig(
        level=0,
        format=(
            "[%(asctime)s - %(filename)s:%(lineno)d]\t\t%(levelname)s\t\t%(message)s"
        ),
    )
    verbosity = int(verbosity)  # because of click choice...
    # check if trc file exists
    if not Path(file).exists():
        raise FileNotFoundError(f"{file} does not exist")

    # open trc file
    with open(str(Path(file)), "rb") as f:
        b_data_trc = f.read()
    if verbosity >= 1:
        logging.info("TRC file read done.")

    # create Micromed
    micromed_io = MicromedIO()
    micromed_io.decode_data_header_packet(b_data_trc)
    markers = dict(
        sorted(micromed_io.micromed_header.markers.items())
    )  # ensure sorted samples
    notes = dict(
        sorted(micromed_io.micromed_header.notes.items())
    )  # ensure sorted samples

    # Create a TCP/IP socket
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    # Bind the socket to the port
    server_address = (address, port)
    if verbosity >= 1:
        logging.info("starting up on %s port %s" % server_address)
    is_not_connected = True

    while is_not_connected:
        try:
            sock.connect(server_address)
            is_not_connected = False
        except Exception as e:
            logging.warning(e)

    # some general variables
    sample_length = (
        micromed_io.micromed_header.nb_of_channels
        * micromed_io.micromed_header.nb_of_bytes
    )
    sfreq = micromed_io.micromed_header.min_sampling_rate
    n_samples_per_packet = int(packet_time * 1e-3 * sfreq)
    eeg_packet_length = sample_length * n_samples_per_packet
    header_address = micromed_io.micromed_header.data_address
    if verbosity >= 1:
        logging.info("Connected!")
        logging.info(
            f"got {len(b_data_trc[header_address:])} bytes - Sending {len(b_data_trc[header_address:]) // sample_length} samples for {len(b_data_trc[header_address:]) / (sample_length * sfreq)} sec"
        )

    # Send the Micromed header data
    tcp_header = mmio_tcp.get_tcp_header(
        mmio_tcp.MicromedPacketType.HEADER, header_address
    )
    if verbosity >= 1:
        logging.info("Sending Micromed header")
    sock.send(tcp_header)
    sock.send(b_data_trc[:header_address])  # send micromed header
    time.sleep(0.1)  # give time for the header to be sent

    start_time = datetime.now()
    current_data_sample = 0
    try:
        while True:
            if (current_data_sample + n_samples_per_packet) / sfreq <= (
                datetime.now() - start_time
            ).total_seconds():
                current_data_sample += n_samples_per_packet

                # check that we are on time
                if (current_data_sample + 5 * n_samples_per_packet) / sfreq <= (
                    datetime.now() - start_time
                ).total_seconds():
                    logging.error(
                        "Critical error: Running out of time. Please increase the packe_size so the emulator can send all the data on time."
                    )
                    return

                # SEND MARKER
                to_rm = []
                for marker_sample, marker_value in markers.items():
                    if marker_sample <= current_data_sample:
                        # send marker
                        if verbosity >= 1:
                            logging.info(f"Sending marker: {marker_value}")

                        tcp_header = mmio_tcp.get_tcp_header(
                            mmio_tcp.MicromedPacketType.MARKER, 6
                        )
                        sock.send(tcp_header)
                        sock.send(
                            mmio_tcp.encode_marker_packet(
                                marker_sample, np.uint16(marker_value)
                            )
                        )  # send micromed header
                        to_rm.append(marker_sample)
                    else:
                        # we can stop cause dict is sorted
                        # this is for time optimization
                        break
                # remove sent markers
                [markers.pop(k) for k in to_rm]

                # SEND NOTE_
                to_rm = []
                for note_sample, note_value in notes.items():
                    if note_sample <= current_data_sample:
                        # send marker
                        if verbosity >= 1:
                            logging.info(f"Sending note: {note_value}")

                        tcp_header = mmio_tcp.get_tcp_header(
                            mmio_tcp.MicromedPacketType.NOTE, 44
                        )
                        sock.send(tcp_header)
                        sock.send(mmio_tcp.encode_note_packet(note_sample, note_value))
                        to_rm.append(note_sample)
                    else:
                        # we can stop cause dict is sorted
                        # this is for time optimization
                        break
                # remove sent markers
                [notes.pop(k) for k in to_rm]

                # SEND DATA
                tcp_header = mmio_tcp.get_tcp_header(
                    mmio_tcp.MicromedPacketType.EEG_DATA, eeg_packet_length
                )
                sock.send(tcp_header)
                data_to_send = b_data_trc[
                    header_address
                    + current_data_sample * (sample_length) : header_address
                    + (current_data_sample + n_samples_per_packet) * sample_length
                ]
                sock.send(data_to_send)

                # debug only
                if verbosity == 2:
                    logging.info(
                        f"sending {len(data_to_send)}bytes with packet_len={eeg_packet_length}; current_ds={current_data_sample}; s={header_address}"
                    )

                # check if no more data
                if (
                    current_data_sample + n_samples_per_packet
                    >= len(b_data_trc[header_address:]) / sample_length
                ):
                    logging.info("TRC file over")
                    break

    except Exception as e:
        logging.error("exception catched: " + str(e))

    finally:
        # Clean up the connection
        logging.info("Closing the server")
        sock.close()


if __name__ == "__main__":
    run()
